/*
Realtime Performance Test Application
Copyright (C) 2005-2014, James Gibbons
Released under the GNU Lesser General Public License version 2.1
See http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html

RTC part based on:
Real Time Clock Driver Test/Example Program
Copyright (C) 1996, Paul Gortmaker.

Released under the GNU General Public License, version 2.

Note: Windows, Mac and Linux platforms currently supported.
*/

#ifndef QT_REALTIMEIO_APP_H
#define QT_REALTIMEIO_APP_H

#include <QtCore>
#include <QtGui>
#include <QtNetwork>
#include <QMainWindow>

namespace Ui {
	class Qt_RealtimeIO_AppClass;
	}

static const int HMAX=300;	// maximum histogram coverage (0.0300 seconds)

#if defined(Q_OS_WIN32) || defined(Q_OS_WIN64)

// define minimum Windows support for NT

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0400
#endif

#ifndef WINVER
#define WINVER _WIN32_WINNT
#endif

#include <windows.h>
#include <mmsystem.h>

//---------------------------------------------------------------------------
// high resolution timer object with inline functions
class THighResTimer
{
private:
	LARGE_INTEGER t0,t1,f;
	double et,last_dt;
	bool na;
protected:
public:
	THighResTimer(void)
	{
		QueryPerformanceFrequency(&f);
		na=f.QuadPart==0;
		HighResReset();
	}

	void HighResReset(void)   // reset timer
	{
		QueryPerformanceCounter(&t0);
		last_dt=0.0;
		et=0.0;
	}

	double HighResTime(void)  // elapsed time in seconds
	{
		double dt;

		if (na) return 0.0;

		QueryPerformanceCounter(&t1);
		dt=t1.QuadPart-t0.QuadPart;
		if (dt<0.0)
		{
			et=et+last_dt/f.QuadPart;
			t0=t1;
			last_dt=0.0;
			return et;
		}
		else
		{
			last_dt=dt;
			return et+dt/f.QuadPart;
		}
	}
};

#endif

#if defined(Q_OS_LINUX) || defined(Q_OS_MAC)

#include <sys/time.h>

//---------------------------------------------------------------------------
// high resolution timer object with inline functions
// only as good as the Linux gettimeofday function (possibly micro-second precision)
class THighResTimer
{
private:
	double sec1;
protected:
public:
	void HighResReset(void)   // reset timer
	{
		struct timeval tv1;
		gettimeofday(&tv1,NULL);
		sec1=tv1.tv_sec+tv1.tv_usec/1000000.0;
	}

	THighResTimer(void)
	{
		HighResReset();
	}

	double HighResTime(void)  // elapsed time in seconds
	{
		struct timeval tv2;
		gettimeofday(&tv2,NULL);
		return (tv2.tv_sec+tv2.tv_usec/1000000.0)-sec1;
	}
};

#endif

//---------------------------------------------------------------------------
// realtime thread class
class RtThread : public QThread
{
	Q_OBJECT

public:
	RtThread(QObject *parent = 0);
	~RtThread();

	// called by main GUI thread to start and stop "realtime" thread
	void startRtThread(QThread::Priority priority, const bool setUnixRunMax=false);
	void stopRtThread();

	// terminates thread
	bool abortRun;

	// setup values
	int clockMode;
	int clockRate;
	QString sendIpAddr;
	bool sendUDP;

	// statistics
	quint64 runCount;
	double runPeriod;
	double maxTime;
	double tt,t1,t2;
	quint64 hSec[HMAX];
	quint64 totalCount;

protected:
	// these are created and released in run() so "realtime" thread owns them
	QTimer *runTimer;
	QUdpSocket *udpSocket;
	int fd;	// linux only, RTC device file

	THighResTimer HRT;

	bool unixRunMax;	// special UNIX max priority enabled

	// run() is called by "realtime" thread
	void run();

	// linux only
	void close_rtc();

protected slots:

	// these are called on timer and udp signals
	void runTimerTick();
	void udpRead();
};

//---------------------------------------------------------------------------
// realtime I/O app
class Qt_RealtimeIO_App : public QMainWindow
{
	Q_OBJECT

public:
	Qt_RealtimeIO_App(QWidget *parent = 0);
	~Qt_RealtimeIO_App();

private:
	Ui::Qt_RealtimeIO_AppClass *ui;

	// the "realtime" thread
	RtThread rtThread;

	// timer used to run GUI display
	QTimer displayTimer;

	void setButtons(bool b);

private slots:
	void runDisplay();
	void on_stopButton_clicked();
	void on_startButton_clicked();
	void on_sendUDP_stateChanged(int);
};

#endif // QT_REALTIMEIO_APP_H
