/*
SQL Performance Test Application
Copyright (C) 2005-2014, James Gibbons
Released under the GNU Lesser General Public License version 2.1
See http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
*/
#ifndef QT_SQL_APP_H
#define QT_SQL_APP_H

#include <QMainWindow>
#include <QtSql>

namespace Ui {
	class Qt_SQL_AppClass;
}

#if defined(Q_OS_WIN32) || defined(Q_OS_WIN64)

// define minimum Windows support for NT

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0400
#endif

#ifndef WINVER
#define WINVER _WIN32_WINNT
#endif

#include <windows.h>

//---------------------------------------------------------------------------
// high resolution timer object with inline functions
class THighResTimer
{
private:
	LARGE_INTEGER t0,t1,f;
	double et,last_dt;
	bool na;
protected:
public:
	THighResTimer(void)
	{
		QueryPerformanceFrequency(&f);
		na=f.QuadPart==0;
		HighResReset();
	}

	void HighResReset(void)   // reset timer
	{
		QueryPerformanceCounter(&t0);
		last_dt=0.0;
		et=0.0;
	}

	double HighResTime(void)  // elapsed time in seconds
	{
		double dt;

		if (na) return 0.0;

		QueryPerformanceCounter(&t1);
		dt=t1.QuadPart-t0.QuadPart;
		if (dt<0.0)
		{
			et=et+last_dt/f.QuadPart;
			t0=t1;
			last_dt=0.0;
			return et;
		}
		else
		{
			last_dt=dt;
			return et+dt/f.QuadPart;
		}
	}
};

#endif

#if defined(Q_OS_LINUX) || defined(Q_OS_MAC)

#include <unistd.h>		// for usleep() use
#include <stdlib.h>		// for rand() use
#include <sys/time.h>

//---------------------------------------------------------------------------
// high resolution timer object with inline functions
// only as good as the Linux gettimeofday function (possibly micro-second precision)
class THighResTimer
{
private:
	double sec1;
protected:
public:
	void HighResReset(void)   // reset timer
	{
		struct timeval tv1;
		gettimeofday(&tv1,NULL);
		sec1=tv1.tv_sec+tv1.tv_usec/1000000.0;
	}

	THighResTimer(void)
	{
		HighResReset();
	}

	double HighResTime(void)  // elapsed time in seconds
	{
		struct timeval tv2;
		gettimeofday(&tv2,NULL);
		return (tv2.tv_sec+tv2.tv_usec/1000000.0)-sec1;
	}
};

#endif

// constants to control various functions

static const int NB_SIZE=16*1024;					// blob size
static const int NR_LIMIT=512;						// records in limited query

class Qt_SQL_App : public QMainWindow
{
	Q_OBJECT

public:
	Qt_SQL_App(QWidget *parent = 0);
	~Qt_SQL_App();

	quint64 runCount,lastRunCount,lastAutoInc;

private:
	Ui::Qt_SQL_AppClass *ui;

	QString appFilePath,appBaseName;

	QTimer displayTimer;
	QTimer idleTimer;
	THighResTimer HRT;

	bool dbOpen;

	void createTable(const QString tableName);

	void openDB();
	void closeDB();

	void resetData();

	int currentMode;

	QByteArray insertBlob;

	QSqlQuery currentQuery;
	QSqlQuery nullQuery;
	QSqlQueryModel model1,model2;

	bool checkExecError(QSqlQuery &query);

	void setControls();

private slots:
	void on_closePB_clicked();
	void on_openPB_clicked();
	void on_dropTable2Button_clicked();
	void on_dropTable1Button_clicked();
	void runQuery();
	void runDisplay();
	void on_emptyTable2Button_clicked();
	void on_emptyTable1Button_clicked();
	void on_copyRecordsButton_toggled(bool);
	void on_insertRecordsButton_toggled(bool);
	void on_sequentialReadButton_toggled(bool);
	void on_noneButton_toggled(bool);
	void on_tabWidget_currentChanged(int);
};

#endif // QT_SQL_APP_H
